"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeGrammarsFile = exports.writeLanguagesFile = exports.extractLanguages = void 0;
const micromatch_1 = __importDefault(require("micromatch"));
const assert_1 = __importDefault(require("assert"));
const utils_1 = require("../utils");
const utils_2 = require("../../src/utils");
const path_1 = __importDefault(require("path"));
const fs_extra_1 = require("fs-extra");
const LANGUAGES = [];
const GRAMMARS = {};
const GRAMMAR_FILES_MAP = new Map();
const DETECTED_EXTENSIONS = new Set();
const DETECTED_FILE_NAME_PATTERNS = new Set();
const DETECTED_FILE_NAMES = new Set();
const DETECTED_FIRST_LINES = new Set();
const EXCLUDED_LANGUAGES = new Set([
    // internal to VS Code
    "search-result",
    // languages without grammars (some third-party extensions can provide them).
    "jsx-tags",
    "juliamarkdown",
]);
async function extractLanguages(contents, packageDir) {
    for (const { id, extensions, filenames, filenamePatterns, firstLine } of contents.contributes?.languages || []) {
        if (EXCLUDED_LANGUAGES.has(id)) {
            continue;
        }
        let language = LANGUAGES.find((l) => l.id === id);
        if (!language) {
            language = {
                id,
                extensions: [],
                fileNames: [],
                fileNamePatterns: [],
                firstLines: [],
            };
            LANGUAGES.push(language);
        }
        language.extensions.push(...(extensions || []));
        language.fileNames.push(...(filenames || []));
        language.fileNamePatterns.push(...(filenamePatterns || []));
        !firstLine || language.firstLines.push(firstLine);
        console.log(`    | 📚 ${id}`);
    }
    for (const { language: languageId, embeddedLanguages, injectTo, path: grammarRelativePath, scopeName, tokenTypes, } of contents.contributes?.grammars || []) {
        if (languageId) {
            if (EXCLUDED_LANGUAGES.has(languageId)) {
                continue;
            }
            switch (scopeName) {
                case "source.cpp.embedded.macro":
                case "text.html.php": {
                    // There are some extensions that define multiple grammars for the same language.
                    // We should skip the "other" grammar to only include the main ones.
                    // https://github.com/microsoft/vscode/issues/127917
                    break;
                }
                default: {
                    const language = LANGUAGES.find((l) => l.id === languageId);
                    assert_1.default(language);
                    assert_1.default(!language.scopeName, "Duplicate scopeName for language: " + languageId);
                    language.scopeName = scopeName;
                    break;
                }
            }
        }
        const grammarFullPath = path_1.default.join(packageDir, grammarRelativePath);
        const grammarFileName = scopeName + path_1.default.extname(grammarFullPath);
        assert_1.default(!GRAMMARS[scopeName] || GRAMMARS[scopeName].grammarFileName === grammarFileName, "Duplicate grammar for: " + scopeName);
        if (GRAMMAR_FILES_MAP.has(grammarFileName)) {
            assert_1.default(GRAMMAR_FILES_MAP.get(grammarFileName) === grammarFullPath);
        }
        else {
            GRAMMAR_FILES_MAP.set(grammarFileName, grammarFullPath);
        }
        const grammarDestination = path_1.default.join(await utils_2.REPO_ROOT, "metadata", "grammars", grammarFileName);
        await fs_extra_1.copy(grammarFullPath, grammarDestination);
        GRAMMARS[scopeName] = {
            injectTo: injectTo || [],
            tokenTypes: tokenTypes || {},
            embeddedLanguages: embeddedLanguages || {},
            grammarFileName,
        };
        console.log(`    | 📄 ${scopeName}`);
    }
}
exports.extractLanguages = extractLanguages;
async function writeLanguagesFile() {
    for (const { extensions, fileNamePatterns, fileNames, firstLines, scopeName } of LANGUAGES) {
        assert_1.default(scopeName);
        sortArray(extensions, DETECTED_EXTENSIONS);
        sortArray(fileNamePatterns, DETECTED_FILE_NAME_PATTERNS);
        sortArray(fileNames, DETECTED_FILE_NAMES);
        sortArray(firstLines, DETECTED_FIRST_LINES);
        fileNamePatterns.forEach((pattern) => {
            assert_1.default(micromatch_1.default.makeRe(pattern));
        });
    }
    LANGUAGES.sort((a, b) => a.id.localeCompare(b.id));
    const destination = path_1.default.join(await utils_2.REPO_ROOT, "metadata", "languages.json");
    await utils_1.writeFormattedJSON(destination, LANGUAGES);
}
exports.writeLanguagesFile = writeLanguagesFile;
async function writeGrammarsFile() {
    for (const scopeName of Object.keys(GRAMMARS)) {
        const grammar = GRAMMARS[scopeName];
        sortArray(grammar.injectTo);
        grammar.embeddedLanguages = sortObject(grammar.embeddedLanguages);
        grammar.tokenTypes = sortObject(grammar.tokenTypes);
    }
    const destination = path_1.default.join(await utils_2.REPO_ROOT, "metadata", "grammars.json");
    await utils_1.writeFormattedJSON(destination, sortObject(GRAMMARS));
}
exports.writeGrammarsFile = writeGrammarsFile;
function sortArray(ar, uniques = new Set()) {
    for (const value of ar) {
        assert_1.default(!uniques.has(value), "duplicate value: " + value);
        uniques.add(value);
    }
    ar.sort();
}
function sortObject(obj) {
    return Object.keys(obj)
        .sort()
        .reduce((acc, key) => {
        acc[key] = obj[key];
        return acc;
    }, {});
}
//# sourceMappingURL=languages.js.map