"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createHighlightingTestSuite = void 0;
const fs_extra_1 = require("fs-extra");
const path_1 = __importDefault(require("path"));
require("jest-specific-snapshot");
const fs_1 = __importDefault(require("fs"));
const theme_wrapper_1 = require("../../src/themes/theme-wrapper");
const requests_1 = require("./workers/requests");
const chalk_1 = __importDefault(require("chalk"));
const trivia_1 = require("../../src/highlighting/trivia");
const utils_1 = require("../../scripts/utils");
function createHighlightingTestSuite(rootDir, callback) {
    const suiteId = path_1.default.basename(rootDir);
    describe(`${suiteId} suite`, () => {
        const requests = new requests_1.RequestWriter(suiteId);
        const themeMetadata = fs_extra_1.readJSONSync(path_1.default.join(__dirname, "../../metadata/themes.json"));
        const originalLevel = chalk_1.default.level;
        beforeAll(() => {
            chalk_1.default.level = 3; // all colors
        });
        afterAll(() => {
            chalk_1.default.level = originalLevel;
        });
        for (const languageId of fs_1.default.readdirSync(rootDir)) {
            if (languageId.endsWith(".test.ts")) {
                continue;
            }
            const languageDir = path_1.default.join(rootDir, languageId);
            describe(`${languageId} language`, () => {
                for (const testId of fs_1.default.readdirSync(languageDir)) {
                    const testDir = path_1.default.join(languageDir, testId);
                    const resultsDir = path_1.default.join(testDir, "results");
                    const discoveredResults = [];
                    describe(`${testId} test`, () => {
                        defineTest(undefined, "#000000");
                        for (const theme of themeMetadata) {
                            defineTest(theme.id, utils_1.getVariationBackground(theme.variation));
                        }
                        function defineTest(themeId, backgroundColor) {
                            it(`${themeId} theme`, async () => {
                                const { sources, lines } = await callback({ testDir, themeId, languageId });
                                const outputFileName = `output.${themeId}.ts`;
                                const outputFullPath = path_1.default.join(resultsDir, outputFileName);
                                discoveredResults.push(outputFileName);
                                const theme = await theme_wrapper_1.ThemeWrapper.create(themeId);
                                const grammar = await theme.loadGrammar(languageId);
                                expect({
                                    lines,
                                    tokens: sources.map((source) => serializeTokens(source, grammar)),
                                }).toMatchSpecificSnapshot(outputFullPath);
                                const screenshotFileName = `screenshot.${themeId}.png`;
                                const screenshotFullPath = path_1.default.join(resultsDir, screenshotFileName);
                                discoveredResults.push(screenshotFileName);
                                requests.add({ destination: screenshotFullPath, lines, backgroundColor });
                            });
                        }
                        afterAll(async () => {
                            const allResults = await fs_extra_1.readdir(resultsDir);
                            await Promise.all(allResults.map(async (result) => {
                                if (!discoveredResults.includes(result)) {
                                    await fs_extra_1.unlink(path_1.default.join(resultsDir, result));
                                }
                            }));
                        });
                    });
                }
            });
            afterAll(async () => {
                await requests.writeToDisk();
            });
        }
    });
}
exports.createHighlightingTestSuite = createHighlightingTestSuite;
function serializeTokens(source, grammar) {
    const lines = trivia_1.splitLines(source);
    const tokens = grammar.tokenizeLines(lines);
    return ("\n" +
        tokens
            .map((line, lineIndex) => {
            const margin = lineIndex.toString().padStart(2, " ") + " │ ";
            if (line.length === 0) {
                return `${margin}none  │         │`;
            }
            return line
                .map(({ line, startColumn, endColumn, foregroundColor }) => {
                expect(line).toEqual(lineIndex);
                return [
                    margin,
                    startColumn.toString().padStart(2),
                    ",",
                    endColumn.toString().padStart(2),
                    " │ ",
                    foregroundColor || "───────",
                    " │ ",
                    lines[lineIndex].substring(startColumn, endColumn),
                ].join("");
            })
                .join("\n");
        })
            .join(`\n───┼───────┼─────────┼──────────────────────────────────────────────────\n`) +
        "\n");
}
//# sourceMappingURL=create-suite.js.map