"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.renderScreenshots = void 0;
const path_1 = __importDefault(require("path"));
const workerpool_1 = __importDefault(require("workerpool"));
const os_1 = __importDefault(require("os"));
const crypto_1 = require("crypto");
const fs_extra_1 = require("fs-extra");
const colors_1 = __importDefault(require("colors"));
const requests_1 = require("./requests");
const cli_progress_1 = __importDefault(require("cli-progress"));
const chalk_1 = __importDefault(require("chalk"));
const CACHE_FILE_PATH = path_1.default.join(os_1.default.tmpdir(), "rendering-reporter-cache.json");
async function renderScreenshots() {
    if (process.env.CI) {
        logMessage("Screenshot rendering was skipped because of CLI flags.");
        return;
    }
    const requestsMap = await requests_1.RequestWriter.deserialize();
    const totalRequests = [...requestsMap.values()].flat().length;
    logMessage(`${totalRequests} screenshots to update.`);
    if (requestsMap.size === 0) {
        return;
    }
    const startTime = new Date();
    const pool = workerpool_1.default.pool(path_1.default.join(__dirname, "worker.js"));
    const cache = (await fs_extra_1.pathExists(CACHE_FILE_PATH)) ? await fs_extra_1.readJson(CACHE_FILE_PATH) : {};
    const multibar = new cli_progress_1.default.MultiBar({
        clearOnComplete: false,
        hideCursor: true,
        format: `[${colors_1.default.gray("{bar}")}] {percentage}% of {suiteId} ({value}/{total})`,
    });
    await Promise.all([...requestsMap].map(async ([suiteId, requests]) => {
        const bar = multibar.create(requests.length, 0, { suiteId });
        await Promise.all(requests.map(async (request) => {
            await execute(request);
            bar.increment(1, { suiteId });
        }));
        bar.stop();
    }));
    multibar.stop();
    await pool.terminate();
    await fs_extra_1.writeJson(CACHE_FILE_PATH, cache);
    const endTime = new Date();
    const seconds = ((endTime.getTime() - startTime.getTime()) / 1000).toFixed(2);
    logMessage(`${totalRequests} screenshots updated in ${seconds} seconds.`);
    async function execute(request) {
        const hash = await crypto_1.createHash("md5").update(JSON.stringify(request)).digest("hex");
        if (await screenshotExists(request, hash)) {
            return;
        }
        await render(request);
        cache[request.destination] = hash;
    }
    async function screenshotExists(options, hash) {
        if (process.argv.includes("--updateSnapshot") || process.argv.includes("-u")) {
            // Jest update snapshots
            return false;
        }
        if (!(await fs_extra_1.pathExists(options.destination))) {
            // Missing image
            return false;
        }
        return hash === cache[options.destination];
    }
    async function render({ lines, destination, backgroundColor }) {
        const options = {
            data: lines.join("\r\n"),
            backgroundColor,
            margin: 10,
            fontFamily: "Monaco",
            type: "png",
        };
        let lastError;
        for (let retry = 0; retry < 3; retry++) {
            try {
                await pool.exec("renderScreenshot", [options, destination]);
                return;
            }
            catch (error) {
                lastError = error;
            }
        }
        throw lastError;
    }
    function logMessage(message) {
        console.log(chalk_1.default.cyanBright(`\n✂  ${message}\n`));
    }
}
exports.renderScreenshots = renderScreenshots;
//# sourceMappingURL=renderer.js.map